#!/bin/python

from __future__ import annotations
# from typing import TYPE_CHECKING
# if TYPE_CHECKING:
    # from .gateway import MBIOGateway

import socket


class MBIOSocket(object):
    def __init__(self, host, port):
        self._host=host
        self._port=port
        self._socket=None
        self._connected=False

    @property
    def host(self):
        return self._host

    def connect(self, timeout=3.0):
        try:
            if not self._connected:
                self._socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._socket.settimeout(timeout)
                self._socket.connect((self._host, self._port))
                # enable non blocking mode
                self._socket.settimeout(0)
                self._connected=True
                return True
            return self._connected
        except:
            pass

    def isConnected(self):
        return self._connected

    def disconnect(self):
        try:
            self._socket.close()
        except:
            pass
        self._socket=None
        self._connected=False

    def write(self, data):
        try:
            self.connect()
            self._socket.send(data)
            return True
        except:
            self.disconnect()

    def read(self):
        try:
            self.connect()
            data = self._socket.recv(4096)
            if not data:
                self.disconnect()
            return data
        except:
            pass


class MBIOSocketString(MBIOSocket):
    def write(self, data):
        return super().write(data.encode())

    def read(self):
        try:
            return super().read().decode()
        except:
            pass


if __name__ == "__main__":
    pass



from .items import Items
from .xmlconfig import XMLConfig
from .mbio import MBIO, MBIONetworkScanner
from .config import MBIOConfig
from .task import MBIOTask
from .linknotifier import MBIOTaskLinkNotifier
# from .gateway import MBIOGateway
from .device import MBIODevice
from .socket import MBIOSocket, MBIOSocketString
from .belimo import MBIODeviceBelimoP22RTH, MBIODeviceBelimoActuator
from .digimatsmartio import MBIODeviceDigimatSIO
from .metzconnect import MBIODeviceMetzConnectMRDO4
from .metzconnect import MBIODeviceMetzConnectMRDI4, MBIODeviceMetzConnectMRDI10
from .metzconnect import MBIODeviceMetzConnectMRAI8, MBIODeviceMetzConnectMRAOP4
from .ebm import MBIODeviceEBM

from .gateway import MBIOGatewayMetzConnectConfigurator



#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .mbio import MBIO

import time
import datetime
import socket
import hashlib

from digimat.lp import PacketManager, LP

from .task import MBIOTask
from .xmlconfig import XMLConfig

from .valuenotifier import MBIOValueNotifier


kLP_ST=20
kLP_MBIO=0xA4

kMBIO_PING=0
kMBIO_PONG=1
kMBIO_SYNC=2
kMBIO_RESTART=3
kMBIO_UPDATEVALUE=10
kMBIO_WRITEVALUE=11
kMBIO_UPDATEVALUEREQ=12
kMBIO_DECLAREVALUE=13
kMBIO_NOTIFYDISABLE=14
kMBIO_AUTH=15
kMBIO_BADAUTH=16
kMBIO_RAZCONFIG=20
kMBIO_BEEP=21
kMBIO_NETSCAN=22

# kST_DUPLICATERSRC=98


class MBIOLinkPacketManager(PacketManager):
    def __init__(self, link: MBIOLink, lid=0):
        super().__init__()
        self._link=link
        self._lid=lid

    @property
    def link(self):
        return self._link

    @property
    def logger(self):
        return self._link.logger

    def dispose(self):
        self.link.disconnect()
        super().dispose()

    def write(self, data):
        return self.link.write(data)

    def manager(self):
        data=self.link.read()
        if data:
            self.receive(data)

    def lp(self, lptype=kLP_MBIO, idRequest=None):
        lp=LP(packetmanager=self)
        if idRequest is None:
            idRequest=self._lid
        lp.create(lptype, idRequest=self._lid)
        return lp


class MBIOLink(object):
    def __init__(self, mbio: MBIO, host, port=5000, lid=0, passkey=None):
        self._mbio=mbio
        self._host=host
        self._port=port
        self._lid=lid
        self._passkey=None
        if passkey:
            try:
                md5=hashlib.md5()
                md5.update(passkey.encode())
                self._passkey=md5.hexdigest()
            except:
                pass
        self._autoDeclareItems=[]
        self._pongReceived=False
        self._protocolVersion=0
        self._socket=None
        self._connected=False
        self._auth=False
        self._timeoutInhibit=3.0
        self._timeoutActivity=0
        self._delayAutoRefresh=0
        self._packetManager=MBIOLinkPacketManager(self, lid)
        self.registerHandlers()
        self.onInit()

    def onInit(self):
        pass

    @property
    def logger(self):
        return self._mbio.logger

    @property
    def packetManager(self):
        return self._packetManager

    def autoDeclareItemsManager(self):
        if self._autoDeclareItems:
            if self._connected and self._pongReceived:
                try:
                    count=16
                    while count>0:
                        count-=1
                        try:
                            data=self._autoDeclareItems.pop(0)
                            item=data['item']
                            skip=data['skip']
                            align=data['align']
                            value=data['value']
                            # self.logger.info('>DECLAREVALUE %s @%d/%d-%d' % (value.key, item, skip, align))
                            lp=self.packetManager.lp()
                            up=lp.up(kMBIO_DECLAREVALUE)
                            up.writeWord(item)  # start index
                            up.writeByte(skip)  # skip count for each group
                            up.writeStrField(value.key)
                            up.writeStrField(value.keyrootpart())
                            up.writeByte(value.flagsAsValue())
                            up.writeByte(align)
                            up.store()
                            lp.send()
                        except:
                            break
                except:
                    self.logger.exception('autoDeclareItemsManager')
                    self._autoDeclareItems=[]

    def manager(self):
        if time.time()>=self._timeoutActivity:
            self.ping()
        self.packetManager.manager()
        self.autoDeclareItemsManager()

    def resetActivityTimeout(self):
        self._timeoutActivity=time.time()+10

    def connect(self):
        try:
            if not self._connected:
                if time.time()>=self._timeoutInhibit:
                    host=self._mbio.resolveIpAliasToIpAddress(self._host)
                    self.logger.info('Opening link %s:%d' % (host, self._port))
                    self._socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    self._socket.settimeout(3)
                    interface=self._mbio.resolveIpAliasToIpAddress(self._mbio.interface)
                    if interface:
                        self.logger.info('Using interface ip %s' % interface)
                        # ifname=(self._interface+'\0').encode('utf-8')
                        # self._socket.setsockopt(socket.SOL_SOCKET, 25, ifname)
                        self._socket.bind((interface, 0))
                    address=(host, self._port)
                    self._socket.connect(address)
                    self._socket.settimeout(0)
                    self._connected=True
                    self._timeoutInhibit=time.time()+5
                    self.logger.info("Link connected to %s:%d" % (host, self._port))
                    self.resetActivityTimeout()
                    self.onConnect()
                    self._mbio.renotifyValues()
        except:
            self.logger.error("Unable to connect link to %s:%d" % (self._host, self._port))
            self._timeoutInhibit=time.time()+5

    def onConnect(self):
        self._auth=False
        self.ping()

    def isConnected(self):
        return self._connected

    def isAuth(self):
        return self._auth

    def disconnect(self):
        try:
            self._socket.close()
            self._socket=None
        except:
            pass

        if self._connected:
            self.logger.warning("Link disconnected from %s:%d" % (self._host, self._port))
            self._connected=False

    def write(self, data):
        try:
            self.connect()
            self._socket.send(data)
            self.resetActivityTimeout()
            return True
        except:
            self.disconnect()
        return False

    def read(self):
        try:
            self.connect()
            data = self._socket.recv(4096)
            if not data:
                self.disconnect()
                return

            self.resetActivityTimeout()
            return data
        except:
            pass

    def registerHandlers(self):
        self.packetManager.addHandler(kLP_MBIO, kMBIO_PONG, self.onPong)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_SYNC, self.onSync)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_RESTART, self.onRestart)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_WRITEVALUE, self.onWriteValue)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_UPDATEVALUEREQ, self.onUpdateValueRequest)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_NOTIFYDISABLE, self.onNotifyDisable)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_AUTH, self.onAuth)
        self.packetManager.addHandler(kLP_MBIO, kMBIO_BADAUTH, self.onBadAuth)

    def ping(self):
        if self._connected:
            self.logger.info('>PING')
            lp=self.packetManager.lp()
            up=lp.up(kMBIO_PING)
            up.store()
            lp.send()

    def md5(self, data):
        try:
            md5=hashlib.md5()
            md5.update(data.encode())
            return md5.hexdigest()
        except:
            pass

    def salt(self):
        try:
            data=str(datetime.datetime.now())
            return self.md5(data)
        except:
            pass

    def RAZCONFIG(self, what=None):
        if self._connected:
            what=what or '*'
            self.logger.warning('>RAZCONFIG(%s)' % what)
            lp=self.packetManager.lp()
            up=lp.up(kMBIO_RAZCONFIG)
            up.writeStr(what)
            up.store()
            if lp.send():
                return True

    def beep(self):
        if self._connected:
            self.logger.debug('>BEEP')
            lp=self.packetManager.lp()
            up=lp.up(kMBIO_BEEP)
            up.store()
            if lp.send():
                return True

    def netscan(self, network=None):
        if self._connected:
            self.logger.debug('>NETSCAN')
            lp=self.packetManager.lp()
            up=lp.up(kMBIO_NETSCAN)
            if network:
                up.writeStr(network)
            up.store()
            if lp.send():
                return True

    def auth(self, salt):
        if self._connected and salt:
            data='%s-%s' % (salt, self._passkey)
            md5=self.md5(data)
            self.logger.warning('>AUTH (salt:%s, key:%s)' % (salt, md5))
            lp=self.packetManager.lp()
            up=lp.up(kMBIO_AUTH)
            up.writeStr(md5)
            up.store()
            result=lp.send()
            if result:
                # Suppose AUTH is successful
                # FIXME: may me enhanced
                self._auth=True
                return True

    def resync(self):
        if self._connected:
            self.logger.info('>RESYNC')
            lp=self.packetManager.lp()
            up=lp.up(kMBIO_SYNC)
            up.store()
            lp.send()

    def updateValue(self, value):
        if value is not None and value.value is not None:
            try:
                # self.logger.debug('>UPDATEVALUE %s' % value)
                value._notifyCount+=1
                lp=self.packetManager.lp()
                up=lp.up(kMBIO_UPDATEVALUE)
                up.writeStr(value.key)
                up.writeFloat(float(value.value))
                up.writeFloat(value.resolution)
                unit=value.unit
                if unit is None:
                    unit=0xff
                up.writeByte(unit)
                up.writeByte(value.flagsAsValue())
                up.store()
                lp.send()
                if len(value.key)>32:
                    self.logger.error('value %s key name too long' % value)
            except:
                self.logger.exception('updateValue')
                self.logger.error(value)

    def onPong(self, up):
        rid=up.readWord()
        protocolRevision=up.readWord()
        maxNbItems=up.readWord()
        itemCount=up.readWord()
        self._delayAutoRefresh=up.readWord()
        self._pongReceived=True
        self.logger.warning('<PONG(id=%u, protocol=%u, %u/%u items, autoRefresh=%us)' %
            (rid, protocolRevision, itemCount, maxNbItems, self._delayAutoRefresh))

    def onSync(self, up):
        self.logger.warning('<SYNC')

    def onAuth(self, up):
        self.logger.warning('<AUTH')
        salt=up.readStr()
        self.auth(salt)
        self.resync()

    def onBadAuth(self, up):
        if self._auth:
            self.logger.error('<BADAUTH')
            self._auth=False

    def onRestart(self, up):
        self.logger.warning('<RESTART')
        self.disconnect()

    def onWriteValue(self, up):
        key=up.readStr()
        value=up.readFloat()
        unit=up.readByte()
        # flags=up.readByte()
        # self.logger.debug('<WRITEVALUE key=%s value=%.02f' % (key, value))

        mbiovalue=self._mbio.value(key)
        if mbiovalue is not None:
            if mbiovalue.isWritable():
                mbiovalue.value=value
                if unit<0xfe:
                    mbiovalue.unit=unit
                # TODO: flags?
            else:
                self.logger.error('<WRITEVALUE received for an unwritable %s' % key)
        else:
            self.logger.warning('<WRITEVALUE unknown value %s' % key)

    def onUpdateValueRequest(self, up):
        key=up.readStr()
        value=self._mbio.value(key)
        if value is not None:
            self.logger.warning('<UPDATEVALUEREQ %s' % key)
            value.enableNotify(True)
            value.notify(force=True)

    def onNotifyDisable(self, up):
        key=up.readStr()
        value=self._mbio.value(key)
        if value is not None:
            self.logger.warning('<NOTIFYDISABLE %s' % key)
            value.enableNotify(False)

    # def onReadItemResponse(self, up):
        # pid=up.readWord()
        # index=up.readWord()
        # value=up.readFloat()
        # unit=up.readByte()
        # item=self._items.get(index)
        # if item:
            # item.value=value
            # item.unit=unit


class MBIOTaskLinkNotifier(MBIOTask):
    def initName(self):
        count=self._parent.tasks.count()
        if count==0:
            return 'station'
        return 'station%d' % count

    def onInit(self):
        self._link=None
        self._notifier=MBIOValueNotifier(self.getMBIO())
        self._iterValues=None

    def onLoad(self, xml: XMLConfig):
        mbio=self.getMBIO()

        host=xml.get('host')
        port=xml.getInt('port', 5000)
        lid=xml.getInt('id', 0)
        passkey=xml.get('passkey')

        if host:
            self._link=MBIOLink(mbio, host, port=port, lid=lid, passkey=passkey)
        else:
            self.logger.error('Unable to create MBIOLink')

        items={}
        for declare in xml.children('declare'):
            item=declare.getInt('item', 0)
            skip=declare.getInt('autoskip', 0)
            align=declare.getInt('autoalign', 0)
            for value in declare.children('values'):
                values=mbio.values(value.get('key'), sort=True)
                for v in values:
                    if v.key not in items:
                        items[v.key]=v
                        self._link._autoDeclareItems.append({'item': item, 'skip': skip, 'align': align, 'value': v})

    def RAZCpuConfig(self, what=None):
        try:
            return self._link.RAZCONFIG(what)
        except:
            pass

    def beep(self):
        return self._link.beep()

    def poweron(self):
        self._link.connect()
        return True

    def poweroff(self):
        self._link.disconnect()
        return True

    def notifyRefreshManager(self):
        count=8
        while count>0:
            count-=1
            if self._iterValues is None:
                try:
                    # reset list
                    self._iterValues=iter(self.getMBIO()._valuesByKey.values())
                except:
                    pass
            try:
                value=next(self._iterValues)
                # self.logger.warning(value)
                value.notifyManager(self._link._delayAutoRefresh)
            except:
                # force reinit list
                self._iterValues=None
                break

    def run(self):
        self._link.manager()
        if self._link.isConnected():
            if self._link.isAuth():
                if not self._link._autoDeclareItems:
                    count=32
                    while count>0:
                        value=self._notifier.get()

                        if value is None:
                            break

                        count-=1
                        # self.logger.debug('NOTIFYUPDATE %s' % value)
                        self.updateValue(value)

                    self.notifyRefreshManager()
                return 0.1
        return 0.5

    def isError(self):
        if super().isError():
            return True
        if not self._link.isConnected():
            return True
        return False

    def updateValue(self, value):
        self._link.updateValue(value)
        value.signalNotified()


if __name__ == "__main__":
    pass



import functools
import http.server
import threading
import time
import os


class WebServer(object):
    def __init__(self, fpath=None, port=8000):
        self._port=port
        self._interface=''
        self._fpath=fpath
        self._thread=None
        self._httpd=None

    def server(self):
        handler=functools.partial(http.server.SimpleHTTPRequestHandler, directory=self._fpath)
        with http.server.ThreadingHTTPServer((self._interface, self._port), handler) as httpd:
            try:
                self._httpd=httpd
                httpd.serve_forever()
            except:
                pass
            httpd.server_close()

        self._httpd=None

    def start(self):
        self._thread=threading.Thread(target=self.server)
        self._thread.daemon=True
        self._thread.start()
        while True:
            time.sleep(1)
            try:
                files=os.listdir(self._fpath)
                if not files:
                    break
            except:
                break

        self.stop()

    def stop(self):
        if self._thread:
            try:
                self._httpd.shutdown()
            except Exception as e:
                print(e)
                pass
            self._thread.join()
            self._thread=None


if __name__ == '__main__':
    ws=WebServer('/tmp/test')
    ws.start()



#!/bin/python

from .device import MBIODevice
from .xmlconfig import XMLConfig

from pymodbus.constants import Endian


class MBIODeviceEBM(MBIODevice):
    def onInit(self):
        self._vendor='EBM'
        self._model='BASE'

        # must return FAN address
        self.setPingHoldingRegister(0xd100)

        self.config.set('enablesource', 'bus')
        self.config.set('speedsource', 'bus')
        self.config.set('controlmode', 'speed')
        self.config.set('maxrpm', 0)
        self.config.set('minspeed', 5.0)
        self.config.set('maxspeed', 100.0)
        self.config.set('ramp')
        self.config.set('ramp0')
        self.config.set('ramp1')
        self.config.set('io1type', 'di')
        self.config.set('io2type', 'ai')
        self.config.set('io3type', 'ao')
        self.config.set('watchdogsource')
        self.config.set('vref', 0)
        self.config.set('iref', 0)
        self.config.set('vout')

        self.value('speed', unit='%')

        self.value('rpm', unit='t/min', resolution=1.0)
        self.value('status', unit=35)
        self.value('warning', unit=35)
        self.value('t1', unit='C')
        self.value('t2', unit='C')
        self.value('t3', unit='C')
        self.valueDigital('clockwise')
        self.value('dci', 'A', resolution=0.2)
        self.value('pow', 'W', resolution=1.0)
        self.value('ene', 'kWh', resolution=0.001)
        self.value('hcnt', unit='h')

        self._timeoutRefreshSlow=0

    def sendWarmReset(self):
        """Reset motor (ack ERROR + reload CONFIG) without communication break"""
        return self.writeRegisters(0xd000, 1)

    def sendColdreset(self):
        """Reset motor (ack ERROR + reload CONFIG) with communication break"""
        return self.writeRegisters(0xd000, 8)

    def onLoad(self, xml: XMLConfig):
        item=xml.child('enable')
        if item:
            self.config.update('enablesource', item.get('source'))
            if self.config.enablesource=='bus':
                self.valueDigital('enable', writable=True)
            else:
                self.valueDigital('enable')

        subitem=xml.child('io1')
        if subitem:
            self.config.update('io1type', subitem.get('type'))

        subitem=xml.child('io2')
        if subitem:
            self.config.update('io2type', subitem.get('type'))

        subitem=xml.child('io3')
        if subitem:
            self.config.update('io3type', subitem.get('type'))

        subitem=xml.child('vout')
        if subitem:
            self.config.update('vout', subitem.getFloat('value', vmin=3.3,  vmax=24.0))

        item=xml.child('speed')
        if item:
            self.config.update('minspeed', item.getFloat('min', vmin=5.0, vmax=50.0))
            self.config.update('maxspeed', item.getFloat('max', vmin=self.config.minspeed, vmax=100.0))
            self.config.update('speedsource', item.get('source'))
            self.config.update('controlmode', item.get('control'))
            self.config.update('ramp', item.getInt('ramp', vmin=0))
            self.config.update('ramp0', item.getInt('ramp0', vmin=0))
            self.config.update('ramp1', item.getInt('ramp1', vmin=0))

            if self.config.controlmode=='sensor':
                # sensor mode
                self.value('ai1', '%', resolution=0.1, commissionable=True)
                self.value('ai2', '%', resolution=0.1, commissionable=True)
                self.value('sp', unit='%', resolution=0.1)
            else:
                # speed mode
                if self.config.speedsource=='bus':
                    value=self.value('sp', unit='%', writable=True, resolution=1, commissionable=True)
                    value.setRange(0, 100)
                else:
                    self.value('ai1', '%', resolution=0.1, commissionable=True)
                    self.value('ai2', '%', resolution=0.1, commissionable=True)
                    self.value('sp', unit='%', resolution=0.1)

            self.config.set('x0', 1)
            self.config.set('y0', 0)
            self.config.set('x1', 10)
            self.config.set('y1', 100)

            subitem=item.child('input')
            if subitem:
                self.config.update('x0', subitem.getFloat('x0', vmin=0, vmax=10))
                self.config.update('x1', subitem.getFloat('x1', vmin=self.config.x0, vmax=10))
                self.config.update('y0', subitem.getFloat('y0', vmin=0, vmax=100))
                self.config.update('y1', subitem.getFloat('y1', vmin=self.config.y0, vmax=100))

            subitem=item.child('watchdog')
            if subitem:
                self.config.set('watchdogsource', subitem.get('source'))
                self.config.set('watchdogsp', subitem.getInt('sp', 0, vmin=0))
                if self.config.watchdogsource=='bus':
                    self.config.set('watchdogdelay', subitem.getInt('delay', 60, vmin=0))
                else:
                    self.config.set('watchdoglevel', subitem.getInt('min', 0, vmin=0))

    def poweron(self):
        # paramerter set 0
        self.writeRegistersIfChanged(0xd105, 0)

        # reload vref, iref
        r=self.readHoldingRegisters(0xd1a0, 2)
        if r:
            self.config.vref=r[0]*20.0/1000.0  # V
            self.config.iref=r[1]*2/1000.0     # A

        # VOUT (fixed power output 3.3..24.0V)
        if self.config.vout is not None:
            v=int((self.config.vout-3.3)/20.7*65536)
            self.writeRegistersIfChanged(0xd16e, v)

        # enable source
        # v6+
        data={'none': 0, 'bus': 1, 'di1-off': 2, 'di2-off': 3, 'di3-off': 4, 'di3': 4, 'di1-on': 5, 'di1': 5, 'di2-on': 6, 'di2': 6}
        self.writeRegistersIfChanged(0xd16a, data.get(self.config.enablesource, 0))

        # speed source
        self.writeRegistersIfChanged(0xd16c, 0)

        # control mode
        data={'speed': 0, 'sensor': 1}
        self.writeRegistersIfChanged(0xd106, data.get(self.config.controlmode, 0))

        # stop enable
        self.writeRegistersIfChanged(0xd112, 1)

        # max/min speed
        v=self.config.maxspeed
        if v is not None:
            self.writeRegistersIfChanged(0xd10e, int(v/100.0*255.0))

        v=self.config.minspeed
        if v is not None:
            self.writeRegistersIfChanged(0xd110, int(v/100.0*255.0))

        r=self.readHoldingRegisters(0xd119, 1)
        if r:
            self.config.update('maxrpm', r[0])

        ramp=self.config.ramp1 or self.config.ramp
        if ramp is not None:
            self.writeRegistersIfChanged(0xd11f, int(float(ramp)/2.5))
        ramp=self.config.ramp0 or self.config.ramp
        if ramp is not None:
            self.writeRegistersIfChanged(0xd120, int(float(ramp)/2.5))

        # input
        if self.config.speedsource != 'bus':
            self.writeRegistersIfChanged(0xd12a, int(self.config.x0/10.0*65535.0))
            self.writeRegistersIfChanged(0xd12c, int(self.config.x1/10.0*65535.0))
            if self.config.controlmode=='speed':
                self.writeRegistersIfChanged(0xd12b, int(self.config.y0/100.0*64000.0))
                self.writeRegistersIfChanged(0xd12d, int(self.config.y1/100.0*64000.0))
            elif self.config.controlmode=='sensor':
                self.writeRegistersIfChanged(0xd12b, int(self.config.y0/100.0*65535.0))
                self.writeRegistersIfChanged(0xd12d, int(self.config.y1/100.0*65535.0))
            elif self.config.controlmode=='openloop':
                self.writeRegistersIfChanged(0xd12b, int(self.config.y0/100.0*65535.0))
                self.writeRegistersIfChanged(0xd12d, int(self.config.y1/100.0*65535.0))

        # output speed monitoring 0-10V
        # self.writeRegistersIfChanged(0xd130, 0)
        self.writeRegistersIfChanged(0xd140, 0)
        self.writeRegistersIfChanged(0xd141, 0)
        self.writeRegistersIfChanged(0xd142, 65535)
        self.writeRegistersIfChanged(0xd143, 65535)

        # speed source
        data={'bus': 1, 'ai1': 0, 'ai2': 2}
        self.writeRegistersIfChanged(0xd101, data.get(self.config.speedsource, 1))

        if self.config.controlmode=='sensor':
            data={'ai1': 0, 'ai2': 1, 'max': 2, 'min': 3, 'mean': 4}
            self.writeRegistersIfChanged(0xd147, data.get(self.config.speedsource, 0))

        # shedding
        self.writeRegistersIfChanged(0xd150, 0)

        # IO1
        data={'di': 0, 'ai': 2}
        self.writeRegistersIfChanged(0xd158, data.get(self.config.io1type, 0))

        # IO2
        data={'di': 0, 'ai': 2}
        self.writeRegistersIfChanged(0xd159, data.get(self.config.io2type, 2))

        # IO3
        data={'di': 0, 'di0': 1, 'di!': 1, 'ao': 3}
        self.writeRegistersIfChanged(0xd15a, data.get(self.config.io3type, 3))

        # watchdog
        if self.config.watchdogsource:
            data={'bus': 1, 'ai1': 2, 'ai2': 2}
            self.writeRegistersIfChanged(0xd15c, data.get(self.config.watchdogsource, 0))
            if self.config.watchdogsource=='bus':
                self.writeRegistersIfChanged(0xd15e,
                    int(self.config.getInt('watchdogdelay', default=15, vmin=10)/100.0))
            elif self.config.watchdogsource in ['ai1', 'ai2']:
                self.writeRegistersIfChanged(0xd15f,
                    self.config.getFloat('watchdoglevel', default=0, vmin=0)/10.0*65535.0)

            v=self.config.getInt('value', default=0, vmin=0, vmax=100)
            if self.config.controlmode=='speed':
                self.writeRegistersIfChanged(0xd15d, int(v/100.0*64000.0))
            else:
                self.writeRegistersIfChanged(0xd15d, int(v/100.0*65535.0))

            # keep rotation direction
            self.writeRegistersIfChanged(0xd15b, 2)
        else:
            self.writeRegistersIfChanged(0xd15c, 0)

        if self.config.controlmode=='sensor':
            # Sensor caracteristics
            encoder=self.encoder(Endian.LITTLE)
            encoder.float32(0)
            encoder.float32(100)
            encoder.writeRegistersIfChanged(0xd160)

        return True

    def poweronsave(self):
        self.sendWarmReset()

    def poweroff(self):
        value=self.values.enable
        if value is not None and value.isWritable():
            value.off()
        value=self.values.sp
        if value is not None and value.isWritable():
            value.set(0)

        self.sync()
        return True

    def refresh(self):
        if self.config.maxrpm is not None:
            r=self.readInputRegisters(0xd010, 1)
            if r:
                try:
                    self.values.rpm.updateValue(r[0]/64000.0*self.config.maxrpm)
                except:
                    pass

        # Max 9 registers with firwmare <6
        r=self.readInputRegisters(0xd011, 9)
        if r:
            self.values.status.updateValue(r[0])
            self.values.warning.updateValue(r[1])
            # self.values.dcv.updateValue(r[2]/255.0*self.config.vref)
            self.values.dci.updateValue(r[3]/255.0*self.config.iref)
            self.values.t1.updateValue(r[4])
            self.values.t2.updateValue(r[5])
            self.values.t3.updateValue(r[6])
            self.values.clockwise.updateValue(r[7])
            self.values.speed.updateValue(r[8]/65535.0*100.0)

        r=self.readInputRegisters(0xd01a, 3)
        if r:
            value=self.values.sp
            if value is not None and not value.isWritable():
                if self.config.controlmode=='speed':
                    value.updateValue(r[1]/64000.0*100.0)
                else:
                    value.updateValue(r[1]/65535*100.0)
            value=self.values.enable
            if value is not None and not value.isWritable():
                value.updateValue(r[2])

        if self.values.ai1 is not None or self.values.ai2 is not None:
            r=self.readInputRegisters(0xd023, 2)
            if r:
                if self.values.ai1 is not None:
                    self.values.ai1.updateValue(r[0]/65535.0*100.0)
                if self.values.ai2 is not None:
                    self.values.ai2.updateValue(r[1]/65535.0*100.0)

        if self.isTimeout(self._timeoutRefreshSlow):
            r=self.readInputRegisters(0xd029, 2)
            if r:
                decoder=self.decoderFromRegisters(r[0:])
                self.values.ene.updateValue(decoder.dword())

            r=self.readInputRegisters(0xd021, 1)
            if r:
                v=r[0]/65535.0*self.config.vref*self.config.iref
                self.values.pow.updateValue(v)

            r=self.readHoldingRegisters(0xd009, 1)
            if r:
                self.values.hcnt.updateValue(r[0])

            self._timeoutRefreshSlow=self.timeout(10)

        return 3.0

    def sync(self):
        value=self.values.enable
        if value is not None and value.isWritable() and value.isPendingSync():
            if self.writeRegisters(0xd00f, value.toReachValue):
                value.clearSyncAndUpdateValue()

        value=self.values.sp
        if value is not None and value.isWritable() and value.isPendingSync():
            vref=65535.0
            if self.config.controlmode=='speed':
                vref=64000.0
            if self.writeRegisters(0xd001, int(value.toReachValue/100.0*vref)):
                value.clearSyncAndUpdateValue()


if __name__ == "__main__":
    pass
